import tensorflow as tf
from wacky_rl import losses

class ActorLoss(losses.WackyLoss):

    def __init__(self, entropy_factor: float = 0.0):
        super().__init__()
        self.entropy_factor = entropy_factor

    def __call__(self, prediction, actions, advantage):

        dist = prediction
        actions = tf.reshape(actions, [-1, len(actions)])
        log_probs = dist.calc_log_probs(actions)
        entropies = dist.calc_entropy(actions)

        if dist.num_actions > 1:
            losses = []
            for i in range(dist.num_actions):
                policy_loss = tf.reduce_mean(tf.math.negative(tf.math.multiply(log_probs[i], advantage)))
                entropy_loss = self.entropy_factor * entropies[i]
                losses.append(policy_loss + entropy_loss)

            loss = tf.reduce_mean(tf.stack(losses))

        else:
            policy_loss = tf.reduce_mean(tf.math.negative(tf.math.multiply(log_probs, advantage)))
            entropy_loss = self.entropy_factor * entropies
            loss = policy_loss + entropy_loss

        return loss


class SoftActorLoss(losses.WackyLoss):

    def __init__(self, entropy_factor: float = 0.0):
        super().__init__()
        self.entropy_factor = entropy_factor

    def __call__(self, prediction, actions, q):

        dist = prediction
        actions = tf.reshape(actions, [-1, len(actions)])
        log_probs = dist.calc_log_probs(actions)
        entropies = dist.calc_entropy(actions)

        if dist.num_actions > 1:
            losses = []
            for i in range(dist.num_actions):
                policy_loss = tf.reduce_mean(tf.squeeze(log_probs[i]) - tf.squeeze(q))
                entropy_loss = self.entropy_factor * entropies[i]
                losses.append(policy_loss + entropy_loss)

            loss = tf.reduce_mean(tf.stack(losses))

        else:
            policy_loss = tf.reduce_mean(tf.squeeze(log_probs) - tf.squeeze(q))
            entropy_loss = self.entropy_factor * entropies
            loss = policy_loss + entropy_loss

        return loss


class PPOActorLoss(losses.WackyLoss):

    def __init__(self, clip_param: float = 0.2, entropy_factor: float = 0.0):
        super().__init__()
        self.clip_param = clip_param
        self.entropy_factor = entropy_factor

    def __call__(self, prediction, actions, old_probs, advantage):

        dist = prediction
        actions = tf.reshape(actions, [-1, len(actions)])
        probs = dist.calc_probs(actions)
        entropies = dist.calc_entropy(actions)

        if dist.num_actions > 1:
            losses = []
            for i in range(dist.num_actions):
                s_1, s_2 = self._calc_surrogates_alternative(tf.stack(probs[i]), tf.stack(old_probs[i]), advantage)
                policy_loss = tf.reduce_mean(tf.math.negative(tf.math.minimum(s_1, s_2)))
                entropy_loss = self.entropy_factor * entropies[i]
                losses.append(policy_loss + entropy_loss)

            loss = tf.reduce_mean(tf.stack(losses))

        else:
            s_1, s_2 = self._calc_surrogates_alternative(tf.stack(probs), tf.stack(old_probs), advantage)
            policy_loss = tf.reduce_mean(tf.math.negative(tf.math.minimum(s_1, s_2)))
            entropy_loss = self.entropy_factor * entropies
            loss = policy_loss + entropy_loss

        return loss

    def _calc_surrogates(self, log_probs, log_old_probs, advantage):

        log_probs = tf.squeeze(log_probs)
        log_old_probs = tf.squeeze(log_old_probs)
        advantage = tf.squeeze(advantage)

        ratios = tf.math.exp(tf.math.log(log_probs + 1e-10) - tf.math.log(log_old_probs + 1e-10))

        sur_1 = tf.math.multiply_no_nan(ratios, advantage)
        sur_2 = tf.math.multiply_no_nan(tf.clip_by_value(ratios, 1.0 - self.clip_param, 1.0 + self.clip_param), advantage)
        return sur_1, sur_2

    def _calc_surrogates_alternative(self, probs, old_probs, advantage):
        sur1 = []
        sur2 = []

        probs = tf.squeeze(probs)
        old_probs = tf.squeeze(old_probs)
        advantage = tf.cast(tf.squeeze(advantage), dtype=tf.float32)

        for pb, t, op in zip(probs, advantage, old_probs):
            t = tf.constant(t)
            op = tf.constant(op)

            ratio = tf.math.divide(pb, op)
            s1 = tf.math.multiply(ratio, t)
            s2 = tf.math.multiply(tf.clip_by_value(ratio, 1.0 - self.clip_param, 1.0 + self.clip_param), t)

            sur1.append(s1)
            sur2.append(s2)

        return tf.stack(sur1), tf.stack(sur2)







import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

from tensorflow.keras import layers

class NormalActionDistributions:

    def __init__(self, num_actions, min_sigma:float = 0.1, max_sigma: float =1.0):

        self.num_actions = num_actions
        self.min_sigma = min_sigma
        self.max_sigma = max_sigma

    def __call__(self, mu_list, sigma_list):
        self.x = [mu_list, sigma_list]
        self.distributions = []
        for i in range(self.num_actions):
            self.distributions.append(
                tfp.distributions.Normal(
                    self._scale_mu(mu_list[i]),
                    self._scale_sigma(sigma_list[i])
                )
            )
        return self

    def _scale_mu(self, mu):
        return tf.squeeze(mu)

    def _scale_sigma(self, sigma):
        sigma = tf.squeeze(tf.clip_by_value(sigma, 0.0, 1.0))
        return sigma * (self.max_sigma - self.min_sigma) + self.min_sigma

    def sample_actions(self):
        actions = []
        for i in range(self.num_actions):
            actions.append(self.distributions[i].sample())
        return tf.math.tanh(tf.stack(actions))

    def mean_actions(self):
        actions = []
        for i in range(self.num_actions):
            actions.append(self.distributions[i].mean())
        return tf.math.tanh(tf.stack(actions))

    def calc_probs(self, actions):
        probs = []
        for i in range(self.num_actions):
            probs.append(self.distributions[i].prob(tf.math.atanh(actions[i])))
        return tf.stack(probs)

    def calc_log_probs(self, actions):
        log_probs = []
        for i in range(self.num_actions):
            log_probs.append(self.distributions[i].log_prob(tf.math.atanh(actions[i])))
        return tf.stack(log_probs)

    def calc_entropy(self, actions):
        entropies = []
        for i in range(self.num_actions):
            entropies.append(self.distributions[i].entropy())
        return tf.stack(entropies)

    def discrete_to_contin(self, actions, num_bins):
        contin_actions = []
        for i in range(self.num_actions):
            contin_actions.append(
                (tf.cast(actions[i], dtype=tf.float32) - (0.5 * (num_bins-1))) / (0.5 * (num_bins - 1))
            )
        return tf.stack(contin_actions)

    def contin_to_discrete(self, actions, num_bins):
        discrete_actions = []
        for i in range(self.num_actions):
            discrete_actions.append(
                (actions[i] * (0.5 * (num_bins - 1))) + (0.5 * (num_bins - 1))
            )
        return tf.cast(tf.stack(discrete_actions), dtype=tf.int32)


class ContinActionLayer(layers.Layer):

    def __init__(
            self,
            num_actions: int,
            mu_activation: str = 'tanh',
            sigma_activation: str = 'sigmoid',
            min_sigma: float = 0.1,
            max_sigma: float = 1.0,
            *args,
            **kwargs
    ):
        super().__init__(**kwargs)

        self._mu_layer = [layers.Dense(1, activation=mu_activation, *args, **kwargs) for _ in range(num_actions)]
        self._sigma_layer = [layers.Dense(1, activation=sigma_activation, *args, **kwargs) for _ in range(num_actions)]
        self.distributions = NormalActionDistributions(num_actions, min_sigma, max_sigma)

    def call(self, inputs, **kwargs):
        mu_list = [mu_l(inputs) for mu_l in self._mu_layer]
        sigma_list = [sigma_l(inputs) for sigma_l in self._sigma_layer]
        return self.distributions(mu_list, sigma_list)



import tensorflow as tf
from tensorflow.keras import layers

from wacky_rl import losses

class WackyModel(tf.keras.Model):

    def __init__(
            self,
            inputs = None,
            outputs = None,
            model_name: str = 'UnnamedWackyModel',
            model_index: int = None,
            *args,
            **kwargs,
    ):

        super().__init__(*args, **kwargs, )

        if inputs is None != outputs is None:
            raise Exception('Both inputs and outputs must be assigned or None.')

        if not inputs is None and not outputs is None:
            self._wacky_layer = [tf.keras.Model(inputs, outputs)]
        elif outputs is None:
            self._wacky_layer = inputs
        else:
            self._wacky_layer = []

        self.model_name = model_name
        self.model_index = model_index

    def add(self, layer):
        if isinstance(layer, list):
            for l in layer: self._wacky_layer.append(l)
        else:
            self._wacky_layer.append(layer)

    def pop(self, index):
        self._wacky_layer.pop(index)

    def nature_network(self, num_units=64, activation='relu'):
        self.add(layers.Dense(num_units, activation=activation))
        self.add(layers.Dense(num_units, activation=activation))

    def compile(
            self,
            optimizer: (str, tf.keras.optimizers.Optimizer) = 'rmsprop',
            loss: (str, losses.WackyLoss, tf.keras.losses.Loss) = 'mse',
            metrics=None,
            loss_weights=None,
            weighted_metrics=None,
            run_eagerly=None,
            steps_per_execution=None,
            **kwargs
    ):

        # Loss Function:
        if isinstance(loss, losses.WackyLoss):
            self._wacky_loss = loss
        else:
            self._wacky_loss = None

        # Taken and modified from Keras:

        from tensorflow.python.keras.engine import compile_utils
        from tensorflow.python.keras.engine import base_layer
        base_layer.keras_api_gauge.get_cell('compile').set(True)
        with self.distribute_strategy.scope():

            # When compiling from an already-serialized model, we do not want to
            # reapply some processing steps (e.g. metric renaming for multi-output
            # models, which have prefixes added for each corresponding output name).
            from_serialized = kwargs.pop('from_serialized', False)

            self._validate_compile(optimizer, metrics, **kwargs)
            self._run_eagerly = run_eagerly

            self.optimizer = self._get_optimizer(optimizer)

            if self._wacky_loss is None:
                self.compiled_loss = compile_utils.LossesContainer(
                    loss, loss_weights, output_names=self.output_names)
            self.compiled_metrics = compile_utils.MetricsContainer(
                metrics, weighted_metrics, output_names=self.output_names,
                from_serialized=from_serialized)

            self._configure_steps_per_execution(steps_per_execution or 1)

            # Initializes attrs that are reset each time `compile` is called.
            self._reset_compile_cache()
            self._is_compiled = True

            if self._wacky_loss is None:
                self.loss = loss or {}  # Backwards compat.

    def _wacky_forward(self, x):
        if not len(self._wacky_layer) == 0:
            for l in self._wacky_layer: x = l(x)
            return x
        else:
            return super().__call__(inputs=x)

    def call(self, inputs, training=False, mask=None, *args, **kwargs):
        return self._wacky_forward(inputs)

    def train_step(self, inputs, *args, **kwargs):

        if self._wacky_loss is None:
            return super().train_step(inputs)

        with tf.GradientTape() as tape:
            x = self(inputs)
            loss = self._wacky_loss(x, *args, **kwargs)
            #loss = self.loss_alpha * self._wacky_loss(x, *args, **kwargs)

        self.optimizer.minimize(loss, self.trainable_variables, tape=tape)
        return loss

    def predict_step(self, data, mask=None, *args, **kwargs):
        return self.call(data, training=False, mask=mask, *args, **kwargs)

    def fit(self, *args, **kwargs):
        if self._wacky_loss is None:
            return super().fit(*args, **kwargs)
        else:
            raise NotImplemented('The fit() method can not be used with a WackyLoss, use train_step() instead.')

    def train_on_batch(self, *args, **kwargs):
        if self._wacky_loss is None:
            return super().fit(*args, **kwargs)
        else:
            raise NotImplemented('The train_on_batch() method can not be used with a WackyLoss, use train_step() instead.')


class WackyDualingModel:

    def __init__(self, model: (WackyModel, tf.keras.model.Model)):

        self.model_1 = model
        self.model_2 = tf.keras.models.clone_model(model)

    @property
    def is_dualing(self):
        return True

    def __call__(self, inputs, training=True, mask=None, *args, **kwargs):
        x_1 = self.model_1(inputs, training, mask)
        x_2 = self.model_2(inputs, training, mask)
        return tf.math.minimum(x_1, x_2)

    def train_step(self, *args, **kwargs):
        loss_1 = self.model_1.train_step(*args, **kwargs)
        loss_2 = self.model_2.train_step(*args, **kwargs)
        return tf.reduce_mean([loss_1, loss_2], 0)

    def predict_step(self, data, mask=None, *args, **kwargs):
        return self(data, training=False, mask=mask, *args, **kwargs)


class TargetUpdate:

    def __init__(self, tau: float = 0.15):
        self.tau = tau

    def __call__(self, model: (WackyModel, tf.keras.model.Model), target: (WackyModel, tf.keras.model.Model)):

        if hasattr(model, 'is_dualing'):
            if model.is_dualing:
                target.model_1 = self._update_target(target.model_1, model.model_1)
                target.model_2 = self._update_target(target.model_2, model.model_2)
                return target
        else:
            return self._update_target(model, target)

    def _update_target(self, model, target):
        weights = model.get_weights()
        target_weights = target.get_weights()

        for i in range(len(target_weights)):
            target_weights[i] = weights[i] * self.tau + target_weights[i] * (1 - self.tau)

        target.set_weights(target_weights)
        return target


class TargetModelWrapper:

    def __init__(self, model, tau: float = 0.15):

        import copy

        self.model = model
        self.target = copy.deepcopy(model)
        self._update_target = TargetUpdate(tau)

    def __call__(self, *args, **kwargs):
        return self.model( *args, **kwargs)

    def train_step(self, *args, **kwargs):
        return self.model.train_step(*args,**kwargs)

    def update_target(self):
        self.target_model = self._update_target(self.model, self.target)



import tensorflow as tf
from wacky_rl import losses


class SoftQLoss(losses.WackyLoss):

    def __init__(self, scale: float = 2.0, gamma: float = 0.99):
        super().__init__()
        self.scale = scale
        self.gamma = gamma

    def __call__(self, prediction, future_vals, rewards, dones, weights=None):

        target = self.scale * tf.squeeze(rewards) + self.gamma * tf.squeeze(future_vals) * tf.squeeze(dones)
        target = tf.reshape(target, [-1,1])

        if not weights is None:
            return weights*tf.keras.losses.MSE(target, prediction)
        return tf.keras.losses.MSE(target, prediction)


import numpy as np
import tensorflow as tf


class ExpectedReturnsCalculator:

    def __init__(self, gamma: float = 0.99, standardize: str = None):

        self.gamma = tf.constant(gamma, dtype=tf.float32)
        self.eps = tf.constant(np.finfo(np.float32).eps.item(), dtype=tf.float32)

        self.static_standardize = False
        self.dynamic_standardize = False

        if standardize == 'static':
            self.static_standardize = True

        elif standardize == 'dynamic':
            self.dynamic_standardize = True

    def __call__(self, rewards):

        sample_len = tf.shape(rewards)[0]
        returns = tf.TensorArray(dtype=tf.float32, size=sample_len)

        rewards = tf.cast(rewards[::-1], dtype=tf.float32)

        discounted_sum = tf.constant(0.0, dtype=tf.float32)
        discounted_sum_shape = discounted_sum.shape

        for i in tf.range(sample_len):

            discounted_sum = rewards[i] + self.gamma * discounted_sum
            discounted_sum.set_shape(discounted_sum_shape)
            returns = returns.write(i, discounted_sum)

        returns = returns.stack()[::-1]

        if self.static_standardize:
            returns = ((returns - tf.math.reduce_mean(returns)) /
                       (tf.math.reduce_std(returns) + self.eps))

        if self.dynamic_standardize:
            self.global_mean = tf.math.reduce_mean(returns) * self.g_rate + self.global_mean * self.one_minus_g_rate
            returns = ((returns - (self.global_mean / 2)) /
                    (tf.math.reduce_std(returns) + self.eps))

        return returns


class LamdaTransformReturns:

    def __init__(
            self,
            gamma: float = 0.99,
            lamda: float = 0.95,
    ):

        self.gamma = gamma
        self.lamda = lamda

    def __call__(self, rewards, dones, values, next_value):
        rewards = tf.squeeze(rewards)
        dones = tf.squeeze(dones)
        values = tf.squeeze(values)
        next_value = tf.squeeze(next_value)
        g = 0
        returns = []
        for i in reversed(range(len(rewards))):
            try:
                delta = rewards[i] + self.gamma * values[i+1] * dones[i] - values[i]
            except:
                delta = rewards[i] + self.gamma * next_value * dones[i] - values[i]
            g = delta + self.gamma * self.lamda * dones[i] * g
            returns.append(g + values[i])

        returns.reverse()
        returns = tf.squeeze(tf.stack(returns))
        adv = returns - values
        adv = (adv - tf.reduce_mean(adv)) / (tf.math.reduce_std(adv) + 1e-10)

        adv = tf.squeeze(tf.stack(adv))
        returns = tf.stack(returns)

        #return tf.reshape(adv, [len(adv),-1]), tf.reshape(returns, [len(returns),-1])
        return adv, returns

class RunningMeanStd(object):
    def __init__(self, epsilon: float = 1e-4, shape= ()):
        """
        Calulates the running mean and std of a data stream
        https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm
        :param epsilon: helps with arithmetic issues
        :param shape: the shape of the data stream's output
        """
        self.mean = np.zeros(shape, np.float64)
        self.var = np.ones(shape, np.float64)
        self.count = epsilon

    def update(self, arr: np.ndarray) -> None:
        batch_mean = np.mean(arr, axis=0)
        batch_var = np.var(arr, axis=0)
        batch_count = arr.shape[0]
        self.update_from_moments(batch_mean, batch_var, batch_count)

    def update_from_moments(self, batch_mean: np.ndarray, batch_var: np.ndarray, batch_count: int) -> None:
        delta = batch_mean - self.mean
        tot_count = self.count + batch_count

        new_mean = self.mean + delta * batch_count / tot_count
        m_a = self.var * self.count
        m_b = batch_var * batch_count
        m_2 = m_a + m_b + np.square(delta) * self.count * batch_count / (self.count + batch_count)
        new_var = m_2 / (self.count + batch_count)

        new_count = batch_count + self.count

        self.mean = new_mean
        self.var = new_var
        self.count = new_count

class GAE:

    def __init__(
            self,
            gamma: float = 0.99,
            lamda: float = 0.95,
    ):

        self.gamma = gamma
        self.lamda = lamda



    def __call__(self, rewards, dones, values, next_value):

        #rewards = tf.squeeze(rewards)


        #rewards = tf.squeeze(rewards).numpy()
        #dones = tf.squeeze(dones).numpy()
        values = tf.squeeze(values).numpy()
        next_value = tf.squeeze(next_value).numpy()

        #returns = tf.TensorArray(size=len(rewards), dtype=tf.float32)
        returns = []
        #advantages = tf.TensorArray(size=len(rewards), dtype=tf.float32)
        advantages = []

        g = 0.0
        for i in reversed(range(len(rewards))):
            try:
                delta = rewards[i] + self.gamma * values[i+1] * dones[i] - values[i]
            except:
                delta = rewards[i] + self.gamma * next_value * dones[i] - values[i]
            g = delta + self.gamma * self.lamda * dones[i] * g
            ret =  g + values[i]
            returns.append(ret)
            advantages.append(g)
            #returns = returns.write(i,ret)
            #advantages = advantages.write(i, g)

        returns.reverse()
        advantages.reverse()


        #adv = advantages.stack()
        #adv_mean = tf.reduce_mean(adv)
        #adv_std = (tf.math.reduce_std(adv) + 1e-10)

        #for i in range(len(adv)):
        #    stand_adv = (adv[i] - adv_mean) / adv_std
        #    advantages = advantages.write(i, stand_adv)

        return tf.stack(advantages), tf.stack(returns)

    def normalize_reward(self, reward: np.ndarray) -> np.ndarray:
        """
        Normalize rewards using this VecNormalize's rewards statistics.
        Calling this method does not update statistics.
        """
        if self.norm_reward:
            reward = np.clip(reward / np.sqrt(self.ret_rms.var + self.epsilon), -self.clip_reward, self.clip_reward)
        return reward


import numpy as np
import tensorflow as tf


def transform_box_to_discrete(dims, act_space):
    n = dims^(np.sum(act_space.shape))
    return n, tf.constant(act_space.high, dtype=tf.float32), tf.constant(act_space.low, dtype=tf.float32)


def transform_discrete_to_box(act_space):
    n = tf.cast(tf.constant(act_space.n), dtype=tf.float32)
    act_shape = (1,)
    return act_shape, n


def contin_act_to_discrete(action, act_shape, highs, lows):
    for dim in act_shape:
        raise NotImplementedError


def discrete_act_to_contin(action, n):
    return tf.cast(tf.cast(action, dtype=tf.float32) * tf.cast(n, dtype=tf.float32), dtype=tf.int32)


class DynamicFactor:

    def __init__(self, tau: float = 1.0, init_factor: float = 1.0):

        self.tau = tau
        self.dynamic_factor = init_factor

    def __call__(self, alpha, max_from_sample):
        self.dynamic_factor = self.tau * max_from_sample + (1 - self.tau) * self.dynamic_factor
        return self.dynamic_factor


class TanhTransformer:

    def __init__(self, alpha: float = 1.0):
        self.alpha = alpha

    def __call__(self, loss):
        return self.alpha * tf.math.tanh(loss)


import tensorflow as tf
import numpy as np

class BaseLogger:

    """
    Creates a summary writer for tensorboard logging.

    Args:
        name (string): Name of the model.
        log_dir (string): Path that indicates which dataset is used.
    """

    def __init__(self, name, log_dir):
        self.name = name

        #logdir = "logs/scalars/" + datetime.now().strftime("%Y%m%d-%H%M%S")
        self.file_writer = tf.summary.create_file_writer(log_dir+"/"+name)
        self.file_writer.set_as_default()


    def log_scalar(self, tag, value, step):
        """Log a scalar variable.

        Args:
            tag (string): Name of the scalar
            value (float): Value of the scalar
            step (int): Training iteration
        """
        tf.summary.scalar(tag, value, step=step)
        self.file_writer.flush()


class TrainingLogger(BaseLogger):
    def __init__(self, name, log_dir):
        super().__init__(name, log_dir)

        self.mean_per_episode_dict = {}

    def log_mean(self, tag, value):

        if tag in set(self.mean_per_episode_dict.keys()):
            self.mean_per_episode_dict[tag].append(value)
        else:
            self.mean_per_episode_dict[tag] = [value]


    def print_status(self, episode):

        print('\nname:', self.name)
        print('Episode:', str(episode))
        for key in self.mean_per_episode_dict.keys():
            if len(self.mean_per_episode_dict[key]) > 1:
                self.mean_per_episode_dict[key] = np.mean(self.mean_per_episode_dict[key])
            else:
                self.mean_per_episode_dict[key] = float(np.squeeze(self.mean_per_episode_dict[key]))
            print('{}: {}'.format(key, self.mean_per_episode_dict[key]))
            self.log_scalar(key, self.mean_per_episode_dict[key], episode)

        self.mean_per_episode_dict.clear()

class StatusPrinter:

    def __init__(self, name):

        self.name = name

        self.mean_per_episode_dict = {}

    def log_mean(self, tag, value):

            if tag in set(self.mean_per_episode_dict.keys()):
                self.mean_per_episode_dict[tag].append(value)
            else:
                self.mean_per_episode_dict[tag] = [value]

    def print_status(self, episode):

        print('\nname:', self.name)
        print('Episode:', str(episode))
        for key in self.mean_per_episode_dict.keys():
            if len(self.mean_per_episode_dict[key]) > 1:
                self.mean_per_episode_dict[key] = np.mean(self.mean_per_episode_dict[key])
            else:
                self.mean_per_episode_dict[key] = float(np.squeeze(self.mean_per_episode_dict[key]))
            print('{}: {}'.format(key, self.mean_per_episode_dict[key]))

        self.mean_per_episode_dict.clear()


class TestingLogger(BaseLogger):
    def __init__(self, name, log_dir):
        super().__init__(name, log_dir)




import tensorflow as tf
from wacky_rl import losses


class SoftValueLoss(losses.WackyLoss):

    def __init__(self):
        super().__init__()

    def __call__(self, prediction, log_probs, q):
        target = tf.reshape(tf.squeeze(q) - tf.squeeze(log_probs), [-1, 1])
        return tf.reduce_mean(tf.keras.losses.MSE(target, prediction))

class MeanSquaredErrorLoss(losses.WackyLoss):

    def __init__(self):
        super().__init__()

    def __call__(self, prediction, returns):
        target = tf.reshape(returns, [-1,1])
        return tf.reduce_mean(tf.keras.losses.MSE(target, prediction))



import random
import numpy as np
import tensorflow as tf

class BufferMemory:

    def __init__(self, maxlen=None):

        self._memory = []
        self.maxlen = maxlen
        self.index_dict = {}
        self._lenght = 0

    @property
    def num_arrays(self):
        return len(self._memory)

    def __len__(self):
        return self._lenght

    def _check_maxlen(self):
        if not self.maxlen is None:
            if self._lenght > self.maxlen:
                self.pop(0)
                self._lenght -= 1

    def _get_index(self, index_or_key):
        if isinstance(index_or_key, str):
            return self.index_dict[index_or_key]
        return index_or_key

    def keys(self):
        return self.index_dict.keys()

    def pop(self, index):
        for elem in self._memory: np.delete(elem, index)

    def pop_array(self, index_or_key):

        index = self._get_index(index_or_key)
        self._memory.pop(index)

        if isinstance(index_or_key, str):
            self.index_dict.pop(index_or_key, None)

        for key in self.index_dict.keys():
            if self.index_dict[key] > index:
                self.index_dict[key] = self.index_dict[key] - 1

    def clear(self):
        self._memory = []
        self.index_dict = {}
        self._lenght = 0

    def __call__(self, to_remember, *arg, **kwargs):

        if isinstance(to_remember, list):
            for i in range(len(to_remember)):
                self._add_item_to_memory(to_remember[i], index=i)

        elif isinstance(to_remember, dict):
            for key in to_remember.keys():
                self._add_item_to_memory(to_remember[key], key=key)

        else:
            self._add_item_to_memory(to_remember, *arg, **kwargs)

        self._lenght += 1
        self._check_maxlen()

    def _add_item_to_memory(self, items_to_remember, key=None, index=None):

        if isinstance(items_to_remember, tf.Tensor):
            items_to_remember = items_to_remember.numpy()

        if not key is None:
            if not key in self.keys():
                self.index_dict[key] = self.num_arrays
            index = self.index_dict[key]

        if not index is None:
            try:
                self._memory[index] = np.concatenate([self._memory[index], np.array([items_to_remember])], axis=0)
            except:
                self._memory.append(np.array([items_to_remember]))

    def __getitem__(self, key_or_index):
        if isinstance(key_or_index, str):
            key_or_index = self._memory[self.index_dict[key_or_index]]
        return self._memory[key_or_index]

    def replay(self, to_tensor=True, indices=None):
        if indices is None:
            mem_list = self._memory
        else:
            mem_list = [elem[indices] for elem in self._memory]
        if to_tensor:
            return [tf.stack(elem) for elem in mem_list]

        return mem_list

    def mini_batches(self, batch_size, num_batches=None, shuffle_batches=False, keys=None):

        if not keys is None:
            item_indices = []
            for key in keys:
                item_indices.append(self.index_dict[key])
        else:
            item_indices = range(self.num_arrays)

        batches = []
        for i in item_indices:
            batches.append(np.array_split(np.copy(self._memory[i]), self._lenght // batch_size))

        batches = list(zip(*batches))

        if shuffle_batches:
            random.shuffle(batches)

        if num_batches is None:
            return batches
        return batches[-num_batches:]

# TODO: ShortTermLongTermBuffer, EpisodicReplayBuffer,


import numpy as np


class GreedyEpsilon:

    def __init__(
            self,
            eps: float = 1.0,
            eps_decay: float = 0.9999,
            eps_min: float = 0.0,
    ):
        self.eps = eps
        self.eps_decay = eps_decay
        self.eps_min = eps_min

    def __call__(self):
        self.eps = max(self.eps * self.eps_decay, self.eps_min)
        if np.random.random() < self.eps:
            return True
        else:
            return False



from wacky_rl.losses._loss_core import WackyLoss

from wacky_rl.losses.actor_losses import ActorLoss
from wacky_rl.losses.actor_losses import SoftActorLoss
from wacky_rl.losses.actor_losses import PPOActorLoss

from wacky_rl.losses.q_losses import SoftQLoss

from wacky_rl.losses.val_losses import SoftValueLoss
from wacky_rl.losses.val_losses import MeanSquaredErrorLoss


import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

from tensorflow.keras import layers

class DiscreteActionDistributions:

    def __init__(self, num_bins, num_actions):

        self.num_bins = num_bins
        self.num_actions = num_actions

    @property
    def x(self):
        return self.distributions

    def __call__(self, x):

        self.distributions = x
        return self

    def sample_actions(self):
        actions = []
        for i in range(self.num_actions):
            actions.append(tf.squeeze(tf.random.categorical(self.distributions[i], num_samples=1), axis=1))
        return tf.stack(actions)

    def mean_actions(self):
        actions = []
        for i in range(self.num_actions):
            actions.append(tf.math.argmax(self.distributions[i], axis=-1))
        return tf.stack(actions)

    def calc_probs(self, actions):
        probs = []
        for i in range(self.num_actions):
            probs.append(
                tf.gather_nd(
                    tf.nn.softmax(self.distributions[i]), tf.stack([np.arange(len(actions[i])), actions[i]], axis=1)
                )
            )
        return tf.stack(probs)

    def calc_log_probs(self, actions):
        log_probs = []
        probs = self.calc_probs(actions)
        for i in range(self.num_actions):
            log_probs.append(tf.math.log(probs[i]))
        return tf.stack(log_probs)

    def calc_entropy(self, actions):
        probs = self.calc_probs(actions)
        entropies =[]
        for i in range(self.num_actions):
            entropies.append(tf.reduce_mean(tf.math.negative(tf.math.multiply(probs[i], tf.math.log(probs[i])))))
        return tf.stack(entropies)

    def actions_to_one_hot(self, actions):
        one_hots = []
        for i in range(self.num_actions):
            one_hots.append(tf.one_hot(actions[i], len(actions[i])))
        return tf.stack(one_hots)

    def discrete_to_contin(self, actions):
        contin_actions = []
        for i in range(self.num_actions):
            contin_actions.append(
                (tf.cast(actions[i], dtype=tf.float32) - (0.5 * (self.num_bins-1))) / (0.5 * (self.num_bins - 1))
            )
        return tf.stack(contin_actions)

    def contin_to_discrete(self, actions):
        discrete_actions = []
        for i in range(self.num_actions):
            discrete_actions.append(
                (actions[i] * (0.5 * (self.num_bins - 1))) + (0.5 * (self.num_bins - 1))
            )
        return tf.cast(tf.stack(discrete_actions), dtype=tf.int32)


class DiscreteActionLayer(layers.Layer):

    def __init__(
            self,
            num_bins,
            num_actions=1,
            activation='softmax',
            *args,
            **kwargs
    ):
        super().__init__(**kwargs)

        self._action_layer = [layers.Dense(num_bins, activation=activation, *args, **kwargs) for _ in range(num_actions)]
        self.distributions = DiscreteActionDistributions(num_bins, num_actions)

    def call(self, inputs, **kwargs):
        action_list = [l(inputs) for l in self._action_layer]
        return self.distributions(action_list)



import tensorflow as tf


def simple_normalize_rewards(rewards):
    return rewards / tf.cast(tf.shape(rewards)[0], dtype=tf.float32)


def forward_discount_rewards(rewards):
    sample_len = tf.shape(rewards)[0]
    placeholder_rewards = tf.TensorArray(dtype=tf.float32, size=sample_len)

    for i in tf.range(sample_len):
        placeholder_rewards = placeholder_rewards.write(i, (rewards[i] / tf.cast(i + 1, dtype=tf.float32)))

    return placeholder_rewards.stack()


def backward_discount_rewards(rewards):

    sample_len = tf.shape(rewards)[0]
    placeholder_rewards = tf.TensorArray(dtype=tf.float32, size=sample_len)
    rewards = rewards[::-1]

    for i in tf.range(sample_len):
        placeholder_rewards = placeholder_rewards.write(i, (rewards[i] / tf.cast(i+1, dtype=tf.float32)))

    return placeholder_rewards.stack()[::-1]


class StaticRewardNormalizer:

    def __init__(self, max_pos_reward, max_neg_reward):
        self.max_pos_reward = max_pos_reward
        self.max_neg_reward = max_neg_reward

    def __call__(self, rewards):
        rewards = tf.where(rewards > 0, rewards / self.max_pos_reward, rewards)
        rewards = tf.where(rewards < 0, rewards / self.max_neg_reward, rewards)
        return rewards


class DynamicRewardNormalizer:

    def __init__(self, max_pos_reward, max_neg_reward, dynamic_rate):
        self.max_pos_reward = max_pos_reward
        self.max_neg_reward = max_neg_reward
        self.dynamic_rate = dynamic_rate

    def __call__(self, rewards):

        rewards = tf.where(rewards > 0, rewards / self.max_pos_reward, rewards)
        rewards = tf.where(rewards < 0, rewards / self.max_neg_reward, rewards)

        min_r = tf.reduce_min(rewards)
        if min_r < 0:
            self.max_neg_reward = self.dynamic_rate * tf.abs(min_r) + (1-self.dynamic_rate) * self.max_neg_reward

        max_r = tf.reduce_max(rewards)
        if max_r > 0:
            self.max_pos_reward = self.dynamic_rate * max_r + (1-self.dynamic_rate) * self.max_pos_reward

        return rewards



#from wacky_rl.layers.actor_layer import DiscreteActionLayer
#from wacky_rl.layers.actor_layer import ContinActionLayer

from wacky_rl.layers.contin_actor_layer import ContinActionLayer
from wacky_rl.layers.discrete_actor_layer import DiscreteActionLayer


'''
absolute value of the magnitude of TD error:
p_t = |\delta_t| + e

stochastic prioritization (probability of being chosen):
P(i) = \frac{p_i^a}{\sum_k p_k^a}

importance sampling weights:
(\frac{1}{N} \times \frac{1}{P(i)})^b

'''
import numpy as np
import tensorflow as tf
import random
from wacky_rl.memory.segment_tree import MinSegmentTree, SumSegmentTree
from wacky_rl.memory import BufferMemory


class PrioritizedReplayBuffer:
    '''
    Based on:
    https://github.com/Curt-Park/rainbow-is-all-you-need/blob/master/03.per.ipynb
    '''

    def __init__(
            self,
            maxlen: int = 10_000,
            alpha: float = 0.6
    ):
        """Initialization."""
        assert alpha >= 0


        self.maxlen = maxlen
        self.max_priority, self.tree_ptr = 1.0, 0
        self.alpha = alpha

        self.memory = BufferMemory(maxlen)

        # capacity must be positive and a power of 2.
        tree_capacity = 1
        while tree_capacity < self.maxlen:
            tree_capacity *= 2

        self.sum_tree = SumSegmentTree(tree_capacity)
        self.min_tree = MinSegmentTree(tree_capacity)

    def remember(self, tensor_list):
        """Store experience and priority."""

        self.memory(tensor_list)

        self.sum_tree[self.tree_ptr] = self.max_priority ** self.alpha
        self.min_tree[self.tree_ptr] = self.max_priority ** self.alpha
        self.tree_ptr = (self.tree_ptr + 1) % self.maxlen

    def sample(self, beta: float = 0.4, batch_size: int = 32):# -> Dict[str, np.ndarray]:
        """Sample a batch of experiences."""
        assert beta > 0

        batch_size = min(batch_size, len(self.memory))

        indices = self._sample_proportional(batch_size)

        mem_list = self.memory.replay(indices=indices)
        weights = tf.cast(tf.stack([self._calculate_weight(i, beta) for i in indices]), dtype=tf.float32)
        return mem_list+[weights,indices]

    def update_priorities(self, indices: list, priorities: np.ndarray):
        """Update priorities of sampled transitions."""
        priorities = priorities.numpy().reshape(-1)
        #print('indices',indices)
        #print('prioritie',priorities)
        assert len(indices) == len(priorities)

        for idx, priority in zip(indices, priorities):
            assert priority > 0
            assert 0 <= idx < len(self.memory)

            self.sum_tree[idx] = priority ** self.alpha
            self.min_tree[idx] = priority ** self.alpha

            self.max_priority = max(self.max_priority, priority)

    def _sample_proportional(self, batch_size):# -> List[int]:
        """Sample indices based on proportions."""
        indices = []
        p_total = self.sum_tree.sum(0, len(self.memory) - 1)
        segment = p_total / batch_size

        for i in range(batch_size):
            a = segment * i
            b = segment * (i + 1)
            upperbound = random.uniform(a, b)
            idx = self.sum_tree.retrieve(upperbound)
            indices.append(idx)

        return indices

    def _calculate_weight(self, idx: int, beta: float):
        """Calculate the weight of the experience at idx."""
        # get max weight
        p_min = self.min_tree.min() / self.sum_tree.sum()
        max_weight = (p_min * len(self.memory)) ** (-beta)

        # calculate weights
        p_sample = self.sum_tree[idx] / self.sum_tree.sum()
        weight = (p_sample * len(self.memory)) ** (-beta)
        weight = weight / max_weight

        return weight


import tensorflow as tf

class ActionScaling:

    def __init__(self, action_space):

        self.low = tf.constant(action_space.low, dtype=tf.float32)
        self.high = tf.constant(action_space.high, dtype=tf.float32)
        self.scale = tf.constant((self.high - self.low) / 2, dtype=tf.float32)
        self.reloc = tf.constant(self.high - self.scale, dtype=tf.float32)

    def scale_to_space(self, action):
        action = action*self.scale + self.reloc
        return tf.clip_by_value(action, self.low, self.high)

    def reverse_scale(self, action):
         action = (action - self.reloc) / self.scale
         return tf.clip_by_value(action, -1.0, 1.0)


#import os
#os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import numpy as np
import tensorflow as tf

from tensorflow.keras.layers import Dense, Input
from tensorflow.keras import Model

from wacky_rl.agents import AgentCore
from wacky_rl.models import WackyModel
from wacky_rl.memory import BufferMemory
from wacky_rl.layers import DiscreteActionLayer, ContinActionLayer
from wacky_rl.losses import PPOActorLoss
from wacky_rl.transform import GAE
from wacky_rl.trainer import Trainer


class TestPPO(AgentCore):

    def __init__(self, env, approximate_contin=False):
        super(TestPPO, self).__init__()

        self.approximate_contin = approximate_contin
        self.memory = BufferMemory()
        self.advantage_and_returns = GAE()

        # Actor:
        num_actions = int(self.decode_space(env.action_space))

        if self.space_is_discrete(env.action_space):
            out_layer = DiscreteActionLayer(num_bins=num_actions)
        elif self.approximate_contin:
            out_layer= DiscreteActionLayer(num_bins=21, num_actions=num_actions)
        else:
            out_layer= ContinActionLayer(num_actions=num_actions)

        self.actor = WackyModel()
        self.actor.nature_network(256)
        self.actor.add(out_layer)
        self.actor.compile(
            optimizer=tf.keras.optimizers.RMSprop(3e-4, clipnorm=0.5),
            loss=PPOActorLoss(entropy_factor=0.001),
        )

        # Critic:
        critic_input = Input(shape=env.observation_space.shape)
        critic_dense = Dense(256, activation='relu')(critic_input)
        critic_dense = Dense(256, activation='relu')(critic_dense)
        critic_out = Dense(1)(critic_dense)
        self.critic = Model(inputs=critic_input, outputs=critic_out)
        self.critic.compile(optimizer='adam', loss='mse')

    def act(self, inputs, act_argmax=False, save_memories=True):

        inputs = tf.expand_dims(tf.squeeze(inputs), 0)
        dist = self.actor(inputs, act_argmax=True)

        if act_argmax:
            actions = dist.mean_actions()
        else:
            actions = dist.sample_actions()

        if save_memories:
            self.memory(actions, key='actions')
            self.memory(dist.calc_probs(actions), key='probs')

        if self.approximate_contin:
            return dist.discrete_to_contin(actions).numpy()
        return actions.numpy()

    def learn(self):

        action, old_probs, states, new_states, rewards, dones = self.memory.replay()
        values = self.critic.predict(states)
        next_value = self.critic.predict(tf.expand_dims(new_states[-1], 0))
        adv, ret = self.advantage_and_returns(rewards, dones, values, next_value)

        for i in range(len(adv)):
            self.memory(adv[i], key='adv')
            self.memory(ret[i], key='ret')

        a_loss_list = []
        c_loss_list = []

        for e in range(3):
            for mini_batch in self.memory.mini_batches(batch_size=32, num_batches=None, shuffle_batches=True):

                action, old_probs, states, new_states, rewards, dones, adv, ret = mini_batch

                adv = tf.squeeze(adv)
                adv = (adv - tf.reduce_mean(adv)) / (tf.math.reduce_std(adv) + 1e-8)

                c_loss = self.critic.train_on_batch(states, ret)
                a_loss = self.actor.train_step(states, action, old_probs, adv)

                a_loss_list.append(tf.reduce_mean(a_loss).numpy())
                c_loss_list.append(tf.reduce_mean(c_loss).numpy())

        self.memory.clear()
        return np.mean(a_loss_list), np.mean(c_loss_list)


def train_ppo():

    import gym
    # env = gym.make('CartPole-v0')
    env = gym.make("LunarLanderContinuous-v2")
    agent = TestPPO(env)

    trainer = Trainer(env, agent)
    trainer.n_step_train(5_000_000, train_on_test=False)
    trainer.test(100)

if __name__ == "__main__":
    train_ppo()



from wacky_rl.transform.value_transformer import transform_box_to_discrete
from wacky_rl.transform.value_transformer import transform_discrete_to_box
from wacky_rl.transform.value_transformer import contin_act_to_discrete
from wacky_rl.transform.value_transformer import discrete_act_to_contin
from wacky_rl.transform.value_transformer import DynamicFactor
from wacky_rl.transform.value_transformer import TanhTransformer

from wacky_rl.transform.returns_transformer import ExpectedReturnsCalculator
from wacky_rl.transform.returns_transformer import LamdaTransformReturns
from wacky_rl.transform.returns_transformer import GAE

from wacky_rl.transform.rewards_transformer import simple_normalize_rewards
from wacky_rl.transform.rewards_transformer import forward_discount_rewards
from wacky_rl.transform.rewards_transformer import backward_discount_rewards
from wacky_rl.transform.rewards_transformer import StaticRewardNormalizer
from wacky_rl.transform.rewards_transformer import DynamicRewardNormalizer


from wacky_rl.models.wacky_model import WackyModel
from wacky_rl.models.wacky_model import TargetUpdate
from wacky_rl.models.wacky_model import TargetModelWrapper
from wacky_rl.models.wacky_model import WackyDualingModel



import numpy as np

class Trainer:

    def __init__(
            self,
            env,
            agent,
    ):

        self.env = env
        self.agent = agent

    def sample_warmup(self, num_episodes):
        for e in range(num_episodes):

            done = False
            obs = self.env.reset()

            while not done:

                action = self.agent.act(obs)
                new_obs, r, done, _ = self.env.step(np.squeeze(action))
                self.agent.memory({
                    'obs': obs,
                    'new_obs': new_obs,
                    'rewards': r,
                    'dones': float(1 - int(done)),
                }
                )
                obs = new_obs

    def episode_train(self, num_episodes):
        for e in range(num_episodes):

            done = False
            obs = self.env.reset()
            reward_list = []

            while not done:

                action = self.agent.act(obs)
                new_obs, r, done, _ = self.env.step(np.squeeze(action))
                reward_list.append(r)
                self.agent.memory({
                    'obs': obs,
                    'new_obs': new_obs,
                    'rewards': r,
                    'dones': float(1 - int(done)),
                }
                )
                obs = new_obs

                if done:
                    a_loss, c_loss = self.agent.learn()
                    print()
                    print('# Episode', e)
                    print('# Sum R:', np.round(np.sum(reward_list), 1))
                    print('# Loss A:', np.round(np.mean(a_loss), 4))
                    print('# Loss C:', np.round(np.mean(c_loss), 4))

    def n_step_train(
            self,
            num_steps,
            n_steps = 2048,
            train_on_test = True,
            render_test = True,
    ):

        train_after = n_steps
        episode_reward_list = []
        s = 0
        while s < num_steps:

            obs = self.env.reset()
            done = False
            reward_list = []

            while not done:

                actions = self.agent.act(obs)
                new_obs, r, done, _ = self.env.step(np.squeeze(actions))

                self.agent.memory({
                    'obs': obs,
                    'new_obs': new_obs,
                    'rewards': r,
                    'dones': float(1 - int(done)),
                }
                )
                reward_list.append(r)
                obs = new_obs
                s += 1

                if done:
                    obs = self.env.reset()
                    episode_reward_list.append(np.sum(reward_list))
                    reward_list = []

            if s >= train_after:
                train_after += n_steps
                a_loss, c_loss = self.agent.learn()
                print()
                print('# Steps', s)
                print('# Mean R:', np.round(np.mean(episode_reward_list), 1))
                print('# Loss A:', np.round(np.mean(a_loss), 4))
                print('# Loss C:', np.round(np.mean(c_loss), 4))
                episode_reward_list = []

                # Test:
                if train_on_test or render_test:
                    done = False
                    while not done:
                        actions = self.agent.act(obs, act_argmax=True, save_memories=train_on_test)
                        new_obs, r, done, _ = self.env.step(np.squeeze(actions))

                        if train_on_test:
                            self.agent.memory({
                                'obs': obs,
                                'new_obs': new_obs,
                                'rewards': r,
                                'dones': float(1 - int(done)),
                            })

                        if render_test:
                            self.env.render()

                        reward_list.append(r)
                        obs = new_obs

                        if done:
                            print('# Test R:', np.round(sum(reward_list), 1))
                            obs = self.env.reset()
                            reward_list = []

        self.env.close()

    def test(self, num_steps):

        for e in range(num_steps):

            done = False
            obs = self.env.reset()
            reward_list = []

            while not done:
                actions = self.agent.act(obs, act_argmax=True, save_memories=False)
                obs, r, done, _ = self.env.step(np.squeeze(actions))
                self.env.render()
                reward_list.append(r)

                if done:
                    print('Epoch', e, '- Sum R:', sum(reward_list))
                    obs = self.env.reset()
                    reward_list = []

        self.env.close()



from wacky_rl import trainer
from wacky_rl import actions
from wacky_rl import agents
from wacky_rl import losses
from wacky_rl import layers
from wacky_rl import memory
from wacky_rl import models
from wacky_rl import transform


from distutils.core import setup
from setuptools import find_packages



setup(
  name = 'wacky-rl',
  packages=find_packages(),
  version = '0.0.4',
  license='MIT',
  description = 'Create custom reinforcement learning agents with wacky-rl.',
  author = 'Maik Schürmann',
  author_email = 'maik.schuermann97@gmail.com',
  url = 'https://github.com/maik97/wacky-rl',
  download_url = 'https://github.com/maik97/wacky-rl/archive/refs/tags/v0.0.4.tar.gz',
  keywords = ['rl', 'actor_critic', 'reinforcement-learning', 'ppo', 'a2c', 'sac', 'dqn', 'keras', 'tensorflow', 'python'],
  install_requires=[
          'tensorflow',
          'tensorflow-probability',
          'gym',
          'numpy',
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.6',
  ],
)




class AgentCore:

    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        return self.act( *args, **kwargs)

    def act(self, inputs, act_argmax=False):
        raise NotImplementedError('When subclassing the `AgentCore` class, you should '
                                  'implement a `act` method.')

    def learn(self, *args, **kwargs):
        raise NotImplementedError('When subclassing the `AgentCore` class, you should '
                                  'implement a `train` method.')

    def decode_space(self, gym_space):

        from gym import spaces

        if isinstance(gym_space, spaces.Box):
            import numpy as np
            return int(np.squeeze(gym_space.shape))

        elif isinstance(gym_space, spaces.Discrete):
            return int(gym_space.n)

        else:
            raise AttributeError('gym_space not understood: {}, use space.Box or space.Discrete'.format(gym_space))

    def space_is_contin(self, gym_space):
        from gym import spaces
        return isinstance(gym_space, spaces.Box)

    def space_is_discrete(self, gym_space):
        from gym import spaces
        return isinstance(gym_space, spaces.Discrete)